package owg.lgm;

import static owg.lgm.plaf.Engine.EN;
import static owg.util.math.Trig.cos;
import static owg.util.math.Trig.sin;

import java.io.BufferedReader;
import java.io.FileNotFoundException;

import owg.lgm.util.FILE;
import owg.lgm.util.Letterbox;

import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.PrintWriter;
import java.lang.reflect.Array;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.nio.FloatBuffer;
import java.nio.ShortBuffer;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import com.jogamp.newt.event.MouseEvent;
import com.jogamp.opengl.GL2ES1;

import owg.lgm.gl.AttribArray;
import owg.lgm.gl.BMFontRenderer;
import owg.lgm.gl.BufferUsage;
import owg.lgm.gl.DataType;
import owg.lgm.gl.DrawCommand;
import owg.lgm.gl.Mesh;
import owg.lgm.gl.PrimitiveType;
import owg.lgm.gl.Tex2D;
import owg.lgm.gl.TextImpl;
import owg.lgm.gl.VertexAttribute;
import owg.lgm.gl.VertexBufferObjectF;
import owg.lgm.gl.DrawCommand.DrawArrays;
import owg.lgm.gl.IndexBufferObject;
import owg.lgm.in.MB;
import owg.lgm.in.VK;
import owg.lgm.plaf.JAVAFILE;
import owg.lgm.plaf.PNGIO;
import owg.lgm.plaf.Sound;
import owg.lgm.room.Room;
import owg.lgm.room.Room.ViewData;
import owg.lgm.room.ActivityChangeException;
import owg.lgm.room.ActivityChangeType;
import owg.lgm.room.RoomData;
import owg.lgm.room.TransitionKind;
import owg.util.color.ARGB;
import owg.util.digitalformat.DigitalFormat;
import owg.util.iorderset.IOrderSet;
import owg.util.math.Clamp;
import owg.util.math.Modulo;
import owg.util.math.Trig;

public class GMLib
{
	//////////////////////////////////////////////////////////////////////////
	// Instances
	//////////////////////////////////////////////////////////////////////////
	
	public static<T extends Entity> T instance_create(double x, double y, Class<T> type)
	{
		try {
			T r = type.getConstructor(double.class, double.class).newInstance(x, y);
			r.createEvent();
			return r;
		} catch (Exception e) {
			throw new RuntimeException("Failed to create instance of "+type+" with arguments "+x+", "+y, e);
		}
	}
	
	/**Returns the canonical individual of the given type, 
	 * i.e. the least recently installed instance of the indicated type which has not yet been uninstalled.*/
	public static<T extends Entity> T instance(Class<T> type)
	{
		return Room.CURRENT.active.first(type);
	}

	/**Returns all installed individuals of the given type, a zero-length array if none.*/
	public static<T> T[] instances(Class<T> type)
	{
		IOrderSet<T> src = Room.CURRENT.active.getInstances(type);
		@SuppressWarnings("unchecked")
		T[] r = (T[]) Array.newInstance(type, src.size());
		src.toArray(r);
		return r;
	}
	
	/**Places all installed individuals of the given type into the given list.*/
	public static<T> void instances(List<? super T> output, Class<T> type)
	{
		Room.CURRENT.active.getInstances(type).addTo(output);
	}
	
	/**Returns the number of installed instances of the given type.*/
	public static int instance_number(Class<? extends Entity> type)
	{
		return Room.CURRENT.active.getCount(type);
	}
	/**Returns whether there are any installed instances of the given type.*/
	public static boolean instance_exists(Class<?> type)
	{
		return Room.CURRENT.active.contains(type);
	}
	
	public static void instance_destroy_all(Class<? extends Entity> type)
	{
		IOrderSet<? extends Entity> set = Room.CURRENT.active.getInstances(type);
		while(!set.isEmpty())
			set.first().destroy();
	}
	
	public static Class<? extends Entity> object(String object)
	{
		Class<? extends Entity> type = EN.gameConfig.objects.get(object);
		if(type == null)
			throw new RuntimeException("No such object type: "+object);
		return type;
	}
	
	public static void instance_deactivate_all(Entity exception)
	{
		Room.CURRENT.deactivateAll(exception);
	}
	public static void instance_deactivate_object(Class<?> type)
	{
		Room.CURRENT.deactivate(type);
	}
	public static void instance_deactivate_object(Entity instance)
	{
		Room.CURRENT.deactivate(instance);	
	}
	public static void instance_activate_all()
	{
		Room.CURRENT.activate(Entity.class);
	}
	public static void instance_activate_object(Class<?> type)
	{
		Room.CURRENT.activate(type);	
	}
	public static void instance_activate_object(Entity instance)
	{
		Room.CURRENT.activate(instance);
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Geometry
	//////////////////////////////////////////////////////////////////////////

	public static double point_direction(double x0, double y0, double x1, double y1)
	{
		return Trig.dir(x0, y0, x1, y1);
	}
	public static double point_distance(double x0, double y0, double x1, double y1)
	{
		return Trig.dist(x0, y0, x1, y1);
	}
	
	public static double lengthdir_x(double len, double dir)
	{
		return Trig.cos(dir)*len;
	}
	public static double lengthdir_y(double len, double dir)
	{
		return -Trig.sin(dir)*len;
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Utility
	//////////////////////////////////////////////////////////////////////////
	
	public static double random()
	{
		return Math.random();
	}
	
	public static double random(double max)
	{
		return max * random();
	}
	
	@SafeVarargs
	public static<T> T choose(T... elements)
	{
		return elements[(int)random(elements.length)];
	}

	public static int choose(int a, int b)
	{
		switch((int)random(2))
		{
		case 0: return a;
		default: return b;
		}
	}
	public static int choose(int a, int b, int c)
	{
		switch((int)random(3))
		{
		case 0: return a;
		case 1: return b;
		default: return c;
		}
	}
	public static int choose(int a, int b, int c, int d)
	{
		switch((int)random(4))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		default: return d;
		}
	}
	public static int choose(int a, int b, int c, int d, int e)
	{
		switch((int)random(5))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		default: return e;
		}
	}
	public static int choose(int a, int b, int c, int d, int e, int f)
	{
		switch((int)random(6))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		default: return f;
		}
	}
	public static int choose(int a, int b, int c, int d, int e, int f, int g)
	{
		switch((int)random(7))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		default: return g;
		}
	}
	public static int choose(int a, int b, int c, int d, int e, int f, int g, int h)
	{
		switch((int)random(8))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		default: return h;
		}
	}
	public static int choose(int a, int b, int c, int d, int e, int f, int g, int h, int i)
	{
		switch((int)random(9))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		default: return i;
		}
	}
	public static int choose(int a, int b, int c, int d, int e, int f, int g, int h, int i, int j)
	{
		switch((int)random(10))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		default: return j;
		}
	}
	public static int choose(int a, int b, int c, int d, int e, int f, int g, int h, int i, int j, int k)
	{
		switch((int)random(11))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		case 9: return j;
		default: return k;
		}
	}
	public static int choose(int a, int b, int c, int d, int e, int f, int g, int h, int i, int j, int k, int l)
	{
		switch((int)random(12))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		case 9: return j;
		case 10: return k;
		default: return l;
		}
	}

	public static double choose(double a, double b)
	{
		switch((int)random(2))
		{
		case 0: return a;
		default: return b;
		}
	}
	public static double choose(double a, double b, double c)
	{
		switch((int)random(3))
		{
		case 0: return a;
		case 1: return b;
		default: return c;
		}
	}
	public static double choose(double a, double b, double c, double d)
	{
		switch((int)random(4))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		default: return d;
		}
	}
	public static double choose(double a, double b, double c, double d, double e)
	{
		switch((int)random(5))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		default: return e;
		}
	}
	public static double choose(double a, double b, double c, double d, double e, double f)
	{
		switch((int)random(6))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		default: return f;
		}
	}
	public static double choose(double a, double b, double c, double d, double e, double f, double g)
	{
		switch((int)random(7))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		default: return g;
		}
	}
	public static double choose(double a, double b, double c, double d, double e, double f, double g, double h)
	{
		switch((int)random(8))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		default: return h;
		}
	}
	public static double choose(double a, double b, double c, double d, double e, double f, double g, double h, double i)
	{
		switch((int)random(9))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		default: return i;
		}
	}
	public static double choose(double a, double b, double c, double d, double e, double f, double g, double h, double i, double j)
	{
		switch((int)random(10))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		default: return j;
		}
	}
	public static double choose(double a, double b, double c, double d, double e, double f, double g, double h, double i, double j, double k)
	{
		switch((int)random(11))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		case 9: return j;
		default: return k;
		}
	}
	public static double choose(double a, double b, double c, double d, double e, double f, double g, double h, double i, double j, double k, double l)
	{
		switch((int)random(12))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		case 9: return j;
		case 10: return k;
		default: return l;
		}
	}

	public static String choose(String a, String b)
	{
		switch((int)random(2))
		{
		case 0: return a;
		default: return b;
		}
	}
	public static String choose(String a, String b, String c)
	{
		switch((int)random(3))
		{
		case 0: return a;
		case 1: return b;
		default: return c;
		}
	}
	public static String choose(String a, String b, String c, String d)
	{
		switch((int)random(4))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		default: return d;
		}
	}
	public static String choose(String a, String b, String c, String d, String e)
	{
		switch((int)random(5))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		default: return e;
		}
	}
	public static String choose(String a, String b, String c, String d, String e, String f)
	{
		switch((int)random(6))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		default: return f;
		}
	}
	public static String choose(String a, String b, String c, String d, String e, String f, String g)
	{
		switch((int)random(7))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		default: return g;
		}
	}
	public static String choose(String a, String b, String c, String d, String e, String f, String g, String h)
	{
		switch((int)random(8))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		default: return h;
		}
	}
	public static String choose(String a, String b, String c, String d, String e, String f, String g, String h, String i)
	{
		switch((int)random(9))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		default: return i;
		}
	}
	public static String choose(String a, String b, String c, String d, String e, String f, String g, String h, String i, String j)
	{
		switch((int)random(10))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		default: return j;
		}
	}
	public static String choose(String a, String b, String c, String d, String e, String f, String g, String h, String i, String j, String k)
	{
		switch((int)random(11))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		case 9: return j;
		default: return k;
		}
	}
	public static String choose(String a, String b, String c, String d, String e, String f, String g, String h, String i, String j, String k, String l)
	{
		switch((int)random(12))
		{
		case 0: return a;
		case 1: return b;
		case 2: return c;
		case 3: return d;
		case 4: return e;
		case 5: return f;
		case 6: return g;
		case 7: return h;
		case 8: return i;
		case 9: return j;
		case 10: return k;
		default: return l;
		}
	}
	
	public static int floor(double value)
	{
		return (int) Math.floor(value);
	}
	
	public static int ceil(double value)
	{
		return (int) Math.ceil(value);
	}
	
	public static int round(double value)
	{
		return (int) Math.round(value);
	}
	
	public static String string_letters(String str)
	{
		StringBuilder r = new StringBuilder(str.length());
		for(int i = 0; i<str.length(); i++)
		{
			char ch = str.charAt(i);
			if(Character.isLetter(ch))
				r.append(ch);
		}
		return r.toString();
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Colors
	//////////////////////////////////////////////////////////////////////////
	
	public static final int c_aqua		= 0xFFFFFF00;
	public static final int c_black 	= 0xFF000000;
	public static final int c_blue 		= 0xFFFF0000;
	public static final int c_cyan 		= 0xFFFFFF00;
	public static final int c_dkgray 	= 0xFF404040;
	public static final int c_fuchsia	= 0xFFFF00FF;
	public static final int c_gray 		= 0xFF808080;
	public static final int c_green 	= 0xFF008000;
	public static final int c_lime		= 0xFF00FF00;
	public static final int c_ltgray 	= 0xFFC0C0C0;
	public static final int c_magenta 	= 0xFFFF00FF;
	public static final int c_maroon	= 0xFF000080;
	public static final int c_navy		= 0xFF800000;
	public static final int c_olive		= 0xFF008080;
	public static final int c_orange	= 0xFF40A0FF;
	public static final int c_purple 	= 0xFF800080;
	public static final int c_red 		= 0xFF0000FF;
	public static final int c_silver 	= 0xFFC0C0C0;
	public static final int c_teal 		= 0xFF808000;
	public static final int c_white 	= 0xFFFFFFFF;
	public static final int c_yellow 	= 0xFF00FFFF;
	
	public static int make_color_rgb(int r, int g, int b)
	{
		return 0xFF000000 | ((b&0xFF)<<16) | ((g&0xFF)<<8) | ((r&0xFF));
	}
	
	public static int color_get_red(int rgb)
	{
		return (rgb&0x000000FF)     ;
	}
	public static int color_get_green(int rgb)
	{
		return (rgb&0x0000FF00)>>> 8;
	}
	public static int color_get_blue(int rgb)
	{
		return (rgb&0x00FF0000)>>>16;
	}
	
	public static int merge_color(int a, int b, double f)
	{
		return (
				0xFF000000 |
				( round( ( ( a&0x000000FF )     ) * (1.0-f) + ( ( b&0x000000FF )     ) * f )    ) |
				( round( ( ( a&0x0000FF00 )>>> 8) * (1.0-f) + ( ( b&0x0000FF00 )>>> 8) * f )<< 8) |
				( round( ( ( a&0x00FF0000 )>>>16) * (1.0-f) + ( ( b&0x00FF0000 )>>>16) * f )<<16)
		);
	}
	
	private static int drawRGB = -1;
	private static double drawAlpha = 1.0;
	private static boolean perspective = false;
	
	public static void draw_set_color(int color)
	{
		drawRGB = color;
	}
	public static int draw_get_color()
	{
		return drawRGB;
	}
	
	public static void draw_set_alpha(double value)
	{
		drawAlpha = Clamp.clamp(value,0,1);
	}
	public static double draw_get_alpha()
	{
		return drawAlpha;
	}
	
	private static byte[] color = new byte[] {(byte) 0xFF, (byte) 0xFF, (byte) 0xFF, (byte) 0xFF};
	private static ByteBuffer colorBuf = ByteBuffer.wrap(color).order(ByteOrder.LITTLE_ENDIAN);
	
	protected static void applyDrawColor()
	{
		colorBuf.putInt(0, drawRGB);
		EN.gl.glColor4ub(color[0], color[1], color[2], (byte) (drawAlpha*0xFF));
	}
	//////////////////////////////////////////////////////////////////////////
	// Fonts
	//////////////////////////////////////////////////////////////////////////
	
	public static final int fa_left = 0;
	public static final int fa_center = 1;
	public static final int fa_right = 2;
	
	public static final int fa_top = 0;
	public static final int fa_middle = 1;
	public static final int fa_bottom = 2;
	
	private static int halign = fa_left, valign = fa_top;
	private static BMFontRenderer font;
	private static HashMap<BMFontRenderer, TextImpl> textCache = new HashMap<>();
	
	public static void draw_set_halign(int align)
	{
		halign = align;
	}
	public static void draw_set_valign(int align)
	{
		valign = align;
	}
	public static void draw_set_font(BMFontRenderer fnt)
	{
		font = fnt==null ? Fonts.get(null) : fnt;
	}
	public static int string_width(CharSequence str)
	{
		return font.getWidth(str, -1);
	}
	public static int string_width_ext(CharSequence str, int lineHeight, int maxWidth)
	{
		return font.getWidth(str, maxWidth);
	}
	public static int string_height(CharSequence str)
	{
		return font.getHeight(str, -1, -1);
	}
	public static int string_height_ext(CharSequence str, int lineHeight, int maxWidth)
	{
		return font.getHeight(str, lineHeight, maxWidth);
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Drawing
	//////////////////////////////////////////////////////////////////////////
	
	private static Mesh rectFill, rectOutline, line, cube;
	private static HashMap<Integer, Mesh> circleCache = new HashMap<>();
	private static VertexBufferObjectF rrBuf;
	private static Mesh rrMesh;
	
	protected static DrawPrimitive drawPrimitive;
	protected static Tex2D drawPrimitiveTexture = null;
	
	protected static D3DPrimitive d3dPrimitive;
	protected static Tex2D d3dPrimitiveTexture = null;
	
	public static FloatBuffer newDirectFloatBuffer(int count)
	{
		return ByteBuffer.allocateDirect(count*4).order(ByteOrder.nativeOrder()).asFloatBuffer();
	}
	
	public static ShortBuffer newDirectShortBuffer(int count)
	{
		return ByteBuffer.allocateDirect(count*2).order(ByteOrder.nativeOrder()).asShortBuffer();
	}
	
	public static void init()
	{
		FloatBuffer cubeVerts = newDirectFloatBuffer(4*8*6);
		cubeVerts.put( 1).put( 1).put( 1).put( 1).put( 0).put( 0).put(0).put(0);
		cubeVerts.put( 1).put(-1).put( 1).put( 1).put( 0).put( 0).put(1).put(0);
		cubeVerts.put( 1).put( 1).put(-1).put( 1).put( 0).put( 0).put(0).put(1);
		cubeVerts.put( 1).put(-1).put(-1).put( 1).put( 0).put( 0).put(1).put(1);
		
		cubeVerts.put( 1).put(-1).put( 1).put( 0).put(-1).put( 0).put(0).put(0);
		cubeVerts.put(-1).put(-1).put( 1).put( 0).put(-1).put( 0).put(1).put(0);
		cubeVerts.put( 1).put(-1).put(-1).put( 0).put(-1).put( 0).put(0).put(1);
		cubeVerts.put(-1).put(-1).put(-1).put( 0).put(-1).put( 0).put(1).put(1);
		
		cubeVerts.put(-1).put(-1).put( 1).put(-1).put( 0).put( 0).put(0).put(0);
		cubeVerts.put(-1).put( 1).put( 1).put(-1).put( 0).put( 0).put(1).put(0);
		cubeVerts.put(-1).put(-1).put(-1).put(-1).put( 0).put( 0).put(0).put(1);
		cubeVerts.put(-1).put( 1).put(-1).put(-1).put( 0).put( 0).put(1).put(1);
		
		cubeVerts.put(-1).put( 1).put( 1).put( 0).put( 1).put( 0).put(0).put(0);
		cubeVerts.put( 1).put( 1).put( 1).put( 0).put( 1).put( 0).put(1).put(0);
		cubeVerts.put(-1).put( 1).put(-1).put( 0).put( 1).put( 0).put(0).put(1);
		cubeVerts.put( 1).put( 1).put(-1).put( 0).put( 1).put( 0).put(1).put(1);
		
		cubeVerts.put(-1).put(-1).put( 1).put( 0).put( 0).put( 1).put(0).put(0);
		cubeVerts.put( 1).put(-1).put( 1).put( 0).put( 0).put( 1).put(1).put(0);
		cubeVerts.put(-1).put( 1).put( 1).put( 0).put( 0).put( 1).put(0).put(1);
		cubeVerts.put( 1).put( 1).put( 1).put( 0).put( 0).put( 1).put(1).put(1);
		
		cubeVerts.put(-1).put( 1).put(-1).put( 0).put( 0).put(-1).put(0).put(0);
		cubeVerts.put( 1).put( 1).put(-1).put( 0).put( 0).put(-1).put(1).put(0);
		cubeVerts.put(-1).put(-1).put(-1).put( 0).put( 0).put(-1).put(0).put(1);
		cubeVerts.put( 1).put(-1).put(-1).put( 0).put( 0).put(-1).put(1).put(1);
		
		cubeVerts.flip();
		VertexBufferObjectF cubeVBuf = new VertexBufferObjectF(EN.gl, cubeVerts, BufferUsage.STATIC_DRAW);
		
		ShortBuffer cubeInds = newDirectShortBuffer(6*6);
		for(int i = 0; i<4*6; i += 4)
			cubeInds.put((short)(i  )).put((short)(i+2)).put((short)(i+1)).put((short)(i+1)).put((short)(i+2)).put((short)(i+3));
		cubeInds.flip();
		IndexBufferObject cubeIBuf = new IndexBufferObject(EN.gl, cubeInds, BufferUsage.STATIC_DRAW);
		
		cube = new Mesh(DrawCommand.drawElements(cubeIBuf, PrimitiveType.TRIANGLES, 6*6, DataType.UNSIGNED_SHORT, 0), 
				new AttribArray(cubeVBuf, VertexAttribute.POSITION, 3, DataType.FLOAT, 4*8, 0),
				new AttribArray(cubeVBuf, VertexAttribute.NORMAL, 3, DataType.FLOAT, 4*8, 4*3),
				new AttribArray(cubeVBuf, VertexAttribute.TEXTURE_COORDINATE_0, 2, DataType.FLOAT, 4*8, 4*6)
		);
		
		FloatBuffer rectVerts = newDirectFloatBuffer(2*4);
		
		rectVerts.put( 0 );
		rectVerts.put( 0 );

		rectVerts.put( 0 );
		rectVerts.put( 1 );

		rectVerts.put( 1 );
		rectVerts.put( 1 );

		rectVerts.put( 1 );
		rectVerts.put( 0 );
		
		rectVerts.flip();
		
		VertexBufferObjectF rectBuf = new VertexBufferObjectF(EN.gl, rectVerts, BufferUsage.STATIC_DRAW);
		rectFill = new Mesh(DrawCommand.drawArrays(PrimitiveType.TRIANGLE_FAN, 0, 4), 
				new AttribArray(rectBuf, VertexAttribute.POSITION, 2, DataType.FLOAT, 2*4, 0),
				new AttribArray(rectBuf, VertexAttribute.TEXTURE_COORDINATE_0, 2, DataType.FLOAT, 2*4, 0));
		
		rectOutline = new Mesh(DrawCommand.drawArrays(PrimitiveType.LINE_LOOP, 0, 4), 
				new AttribArray(rectBuf, VertexAttribute.POSITION, 2, DataType.FLOAT, 2*4, 0));
		
		FloatBuffer lineVerts = newDirectFloatBuffer(4);
		lineVerts.put(0).put(0).put(1).put(0);
		lineVerts.flip();
		VertexBufferObjectF lineBuf = new VertexBufferObjectF(EN.gl, lineVerts, BufferUsage.STATIC_DRAW);
		line = new Mesh(DrawCommand.drawArrays(PrimitiveType.LINES, 0, 2), 
				new AttribArray(lineBuf, VertexAttribute.POSITION, 2, DataType.FLOAT, 2*4, 0));
		
    	
		int vCount = 18;
		FloatBuffer verts = newDirectFloatBuffer(2*vCount);
		rrBuf = new VertexBufferObjectF(EN.gl, verts, BufferUsage.DYNAMIC_DRAW);
		rrMesh = new Mesh(DrawCommand.drawArrays(PrimitiveType.TRIANGLE_FAN, 0, vCount), 
				new AttribArray(rrBuf, VertexAttribute.POSITION, 2, DataType.FLOAT, 4*2, 0));
		rrMesh.autoDestroy(rrBuf);
		
		drawPrimitive = new DrawPrimitive(EN.gl);
		d3dPrimitive = new D3DPrimitive(EN.gl);
	}
	
	public static void draw_background(Background b, double x, double y)
	{
		draw_background_ext(b, x, y, 1, 1, 0, -1, 1);
	}
	public static void draw_background_ext(Background b, double x, double y, double xScale, double yScale, double angle, int blend, double alpha)
	{
		Tex2D texture = b.tex;
		
		float x0 = (float) Math.rint(x);
		float y0 = (float) Math.rint(y);
		
		EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
		EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, texture.getHandle());
		colorBuf.putInt(0, blend);
		EN.gl.glColor4ub(color[0], color[1], color[2], (byte) (Clamp.clamp(alpha,0,1)*0xFF));

		EN.gl.glPushMatrix();
		EN.gl.glTranslatef(x0, y0, 0);
		EN.gl.glRotatef((float)angle, 0, 0, -1);
		EN.gl.glScalef(b.width*(float)xScale, b.height*(float)yScale, 1);
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glTranslatef(texture.tx0, texture.ty0, 0);
		EN.gl.glScalef(texture.tw, texture.th, 1);
		
		rectFill.render(EN.gl);
		
		EN.gl.glLoadIdentity();
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
		EN.gl.glPopMatrix();
	}
	
	public static void draw_background_tiled(Background b, double x, double y)
	{
		draw_background_tiled_ext(b, x, y, true, true, c_white, 1.0);
	}
	public static void draw_background_tiled_ext(Background b, double x, double y, boolean tileX, boolean tileY, int blend, double alpha)
	{
		colorBuf.putInt(0, blend);
		EN.gl.glColor4ub(color[0], color[1], color[2], (byte) (Clamp.clamp(alpha,0,1)*0xFF));

		EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
		EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, b.tex.getHandle());
		
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glTranslatef(b.tex.tx0, b.tex.ty0, 0);
		EN.gl.glScalef(b.tex.tw, b.tex.th, 1);
		
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);

		EN.gl.glPushMatrix();
		
		float baseX = (float) Math.rint(x);
		float baseY = (float) Math.rint(y);
		int vi = Room.CURRENT.getViewIndex();
		int xview, yview, wview, hview;
		if(vi >= 0) {
			xview = Room.CURRENT.views[vi].xview;
			yview = Room.CURRENT.views[vi].yview;
			wview = Room.CURRENT.views[vi].wview;
			hview = Room.CURRENT.views[vi].hview;
		} else {
			xview = 0;
			yview = 0;
			wview = room_width();
			hview = room_height();
		}
		if(tileX)
			baseX = Modulo.mod(baseX, xview-b.width, xview);
		if(tileY)
			baseY = Modulo.mod(baseY, yview-b.height, yview);
		EN.gl.glTranslatef(baseX, baseY, 0);
		EN.gl.glScalef(b.width, b.height, 1);
		if(tileX)
		{
			for(int i = wview/b.width+1; i >= 0; i--)
			{
				draw_background_column(b, tileY, hview);
				EN.gl.glTranslatef(1, 0, 0);
			}
		}
		else
			draw_background_column(b, tileY, hview);
		
		EN.gl.glPopMatrix();
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glLoadIdentity();
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
	}

	protected static void draw_background_column(Background bg, boolean tileY, int hview) 
	{
		if(tileY)
		{
			EN.gl.glPushMatrix();
			for(int i = hview/bg.height+1; i >= 0; i--)
			{
				EN.nwRect.render(EN.gl);
				EN.gl.glTranslatef(0, 1, 0);
			}
			EN.gl.glPopMatrix();
		}
		else
			EN.nwRect.render(EN.gl);
	}

	public static void draw_background_stretched(Background b, double x, double y, double w, double h)
	{
		draw_background_stretched_ext(b, x, y, w, h, c_white, 1.0);
	}
	
	public static void draw_background_stretched_ext(Background b, double x, double y, double w, double h, int blend, double alpha)
	{
		colorBuf.putInt(0, blend);
		EN.gl.glColor4ub(color[0], color[1], color[2], (byte) (Clamp.clamp(alpha,0,1)*0xFF));

		EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
		EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, b.tex.getHandle());
		
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glTranslatef(b.tex.tx0, b.tex.ty0, 0);
		EN.gl.glScalef(b.tex.tw, b.tex.th, 1);
		
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);

		EN.gl.glPushMatrix();
		
		EN.gl.glTranslatef((float)Math.rint(x), (float)Math.rint(y), 1.0f);
		EN.gl.glScalef((float)w, (float)h, 1);
		EN.nwRect.render(EN.gl);
			
		EN.gl.glPopMatrix();
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glLoadIdentity();
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
	}

	public static void draw_sprite(Sprite sprite, double image_index, double x, double y)
	{
		Tex2D texture = sprite.textures[Modulo.mod((int)Math.floor(image_index), sprite.number)];
		
		float x0 = (float) (Math.rint(x) -sprite.get_xoffset());
		float y0 = (float) (Math.rint(y) -sprite.get_yoffset());

		EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
		EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, texture.getHandle());
		EN.gl.glColor4f(1,1,1,1);

		EN.gl.glPushMatrix();
		EN.gl.glTranslatef(x0, y0, 0);
		EN.gl.glScalef(sprite.width, sprite.height, 1);
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glTranslatef(texture.tx0, texture.ty0, 0);
		EN.gl.glScalef(texture.tw, texture.th, 1);
		
		rectFill.render(EN.gl);

		EN.gl.glLoadIdentity();
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
		EN.gl.glPopMatrix();
	}
	
	public static void draw_sprite_ext(Sprite sprite, double image_index, double x, double y, double image_xscale, double image_yscale, double image_angle, int image_blend, double image_alpha)
	{
		double lx0 = -image_xscale*sprite.get_xoffset();
		double ly0 = -image_yscale*sprite.get_yoffset();
		
		double c = Trig.cos(image_angle);
		double s = Trig.sin(image_angle);
		
		Tex2D texture = sprite.textures[Modulo.mod((int)Math.floor(image_index), sprite.number)];
		
		float x0 = (float) (Math.rint(x) + c * lx0 + s * ly0);
		float y0 = (float) (Math.rint(y) - s * lx0 + c * ly0);
		
		EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
		EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, texture.getHandle());
		colorBuf.putInt(0, image_blend);
		EN.gl.glColor4ub(color[0], color[1], color[2], (byte) (Clamp.clamp(image_alpha,0,1)*0xFF));

		EN.gl.glPushMatrix();
		EN.gl.glTranslatef(x0, y0, 0);
		EN.gl.glRotatef((float)image_angle, 0, 0, -1);
		EN.gl.glScalef(sprite.width*(float)image_xscale, sprite.height*(float)image_yscale, 1);
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glTranslatef(texture.tx0, texture.ty0, 0);
		EN.gl.glScalef(texture.tw, texture.th, 1);
		
		rectFill.render(EN.gl);
		
		EN.gl.glLoadIdentity();
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
		EN.gl.glPopMatrix();
	}

    public static void draw_rectangle(double x0, double y0, double x1, double y1, boolean outline)
	{
    	float off = outline ? 0.5f : 0.0f;
    	
    	applyDrawColor();
    	EN.gl.glDisable(GL2ES1.GL_TEXTURE_2D);
    	EN.gl.glPushMatrix();
    	EN.gl.glTranslatef((float)x0+off, (float)y0+off, 0);
    	EN.gl.glScalef((float)(x1-x0), (float)(y1-y0), 1);
    	
    	(outline?rectOutline:rectFill).render(EN.gl);
		EN.gl.glPopMatrix();
	}

    public static void draw_roundrect(double x0, double y0, double x1, double y1, boolean outline)
    {
    	float hr = Math.min(8, (float)(x1-x0)/2.0f);
    	float vr = Math.min(8, (float)(y1-y0)/2.0f);
    	float off = outline ? 0.5f : 0.0f;
    	
    	float fx0 = (float) (x0+hr+off);
    	float fy0 = (float) (y0+vr+off);
    	float fx1 = (float) (x1-hr+off);
    	float fy1 = (float) (y1-vr+off);
    	
		PrimitiveType mode;
		int offset;
		if(outline) {
			mode = PrimitiveType.LINE_STRIP;
			offset = 1;
		} else {
			mode = PrimitiveType.TRIANGLE_FAN;
			offset = 0;
		}
    	
		int n = 4;
		FloatBuffer verts = rrBuf.getData();
		verts.put((fx0+fx1)/2.0f).put((fy0+fy1)/2.0f);
		
		float[] c = {hr  , hr*0.866f, hr*0.5f  , 0.0f};
		float[] s = {0.0f, vr*0.5f  , vr*0.866f, vr  };
		
		for(int i = 0; i<n; i++)
			verts.put(fx1+c[i]).put(fy0-s[i]);
		for(int i = n-1; i>=0; i--)
			verts.put(fx0-c[i]).put(fy0-s[i]);
		for(int i = 0; i<n; i++)
			verts.put(fx0-c[i]).put(fy1+s[i]);
		for(int i = n-1; i>=0; i--)
			verts.put(fx1+c[i]).put(fy1+s[i]);
		
		verts.put(fx1+hr).put(fy0);
		verts.flip();

		rrBuf.update();
		
		DrawArrays cmd = (DrawArrays)rrMesh.drawCommand;
		cmd.first = offset;
		cmd.count = 4*4+2-offset;
		cmd.mode = mode;
		
    	applyDrawColor();
    	EN.gl.glDisable(GL2ES1.GL_TEXTURE_2D);
    	rrMesh.render(EN.gl);
    }
    
    private static float[] mm = new float[16];
	
    public static void draw_line(double x0, double y0, double x1, double y1)
	{
		applyDrawColor();
    	EN.gl.glDisable(GL2ES1.GL_TEXTURE_2D);
		EN.gl.glPushMatrix();
		mm[ 0] = (float)(x1-x0);mm[ 1] = (float)(y1-y0);mm[ 2] = 0; mm[ 3] = 0;
		mm[ 4] = 0;				mm[ 5] = 1;				mm[ 6] = 0; mm[ 7] = 0; 
		mm[ 8] = 0;				mm[ 9] = 0;				mm[10] = 1; mm[11] = 0;
		mm[12] = (float)x0+0.5f; 	mm[13] = (float)y0+0.5f; 	mm[14] = 0; mm[15] = 1;
		EN.gl.glMultMatrixf(mm, 0);
    	line.render(EN.gl);
    	EN.gl.glPopMatrix();
	}
    
    protected static int circleVerts(double r)
    {
    	return 4*(int)((r)/16)+8;
    }

    public static void draw_circle(double x, double y, double r, boolean outline)
	{
    	if(r<=0)
    		return;
		int n = circleVerts(r);
		Mesh cMesh = circleCache.get(n);
		int vCount = n + 2;
		
		PrimitiveType mode;
		int offset;
		if(outline) {
			mode = PrimitiveType.LINE_STRIP;
			offset = 1;
		} else {
			mode = PrimitiveType.TRIANGLE_FAN;
			offset = 0;
		}
		
		if(cMesh == null)
		{
			FloatBuffer verts = newDirectFloatBuffer(2*vCount);
			verts.put((float) 0).put((float) 0);
			
			for(int i = 0; i<=n; i++)
			{
				double a = i * 360.0/n;
				verts.put((float)Trig.cos(a)).put((float)Trig.sin(a));
			}
			verts.flip();
			VertexBufferObjectF cBuf = new VertexBufferObjectF(EN.gl, verts, BufferUsage.STATIC_DRAW);
			cMesh = new Mesh(DrawCommand.drawArrays(mode, offset, vCount), 
					new AttribArray(cBuf, VertexAttribute.POSITION, 2, DataType.FLOAT, 4*2, 0));
			cMesh.autoDestroy(cBuf);
			circleCache.put(n, cMesh);
		}
		else
		{
			DrawArrays cmd = ((DrawArrays)cMesh.drawCommand);
			cmd.mode = mode;
			cmd.first = offset;
			cmd.count = vCount-offset;
		}

    	if(outline) {
    		x+=0.5;
    		y+=0.5;
    	}
    	applyDrawColor();
    	EN.gl.glDisable(GL2ES1.GL_TEXTURE_2D);
    	EN.gl.glPushMatrix();
    	EN.gl.glTranslatef((float)x, (float)y, 0);
    	EN.gl.glScalef((float)r, (float)r, 1);
    	cMesh.render(EN.gl);
    	EN.gl.glPopMatrix();
	}

    public static void draw_circle_color(double x, double y, double r, int innerColor, int outerColor, boolean outline)
    {
    	if(outline) {
    		draw_primitive_begin(PrimitiveType.LINE_STRIP);
    		x+=0.5;
    		y+=0.5;
    	} else {
    		draw_primitive_begin(PrimitiveType.TRIANGLE_FAN);
    		draw_vertex_color(x, y, innerColor, drawAlpha);
    	}
		int n = circleVerts(r);
		for(int i = 0; i<=n; i++)
			draw_vertex_color(x+r*cos(i*360.0/n), y-r*sin(i*360.0/n), outerColor, drawAlpha);
		draw_primitive_end();
    }
    
    public static void draw_ellipse_color(double x0, double y0, double x1, double y1, int innerColor, int outerColor, boolean outline)
    {
    	double xm = (x0+x1)/2;
    	double ym = (y0+y1)/2;
    	double xr = x1-xm;
    	double yr = y1-ym;
    	if(outline) {
    		draw_primitive_begin(PrimitiveType.LINE_STRIP);
    		xm+=0.5;
    		ym+=0.5;
    	} else {
    		draw_primitive_begin(PrimitiveType.TRIANGLE_FAN);
    		draw_vertex_color(xm, ym, innerColor, drawAlpha);
    	}
		int n = circleVerts((xr+yr)/2);
		for(int i = 0; i<=n; i++)
			draw_vertex_color(xm+xr*cos(i*360.0/n), ym-yr*sin(i*360.0/n), outerColor, drawAlpha);
		draw_primitive_end();
    }
    
	public static void draw_healthbar(double x0,double y0,double x1,double y1,double amount, int backcol, int mincol, int maxcol, int anchor, boolean showback, int border)
	{
		amount /= 100;
		if(showback)
		{
			draw_set_color(backcol);
			draw_rectangle(x0-border, y0-border, x1+border, y1+border, false);
		}
		draw_set_color(ARGB.mix(mincol, maxcol, (float) amount));
		double w = x1-x0;
		double h = y1-y0;
		if(anchor == 0)
			draw_rectangle(x0, y0, x0+w*amount, y1, false);
		else if(anchor == 1)
			draw_rectangle(x1-w*amount, y0, x1, y1, false);
		else if(anchor == 2)
			draw_rectangle(x0, y0, x1, y0+h*amount, false);
		else if(anchor == 3)
			draw_rectangle(x0, y1-h*amount, x1, y1, false);
	}
	
	public static void draw_primitive_begin(PrimitiveType kind)
	{
		drawPrimitiveTexture = null;
		drawPrimitive.begin(kind);
	}
	
	public static void draw_primitive_begin_texture(PrimitiveType kind, Tex2D texture)
	{
		drawPrimitiveTexture = texture;
		drawPrimitive.begin(kind);
	}
	
	public static void draw_vertex(double x, double y)
	{
		drawPrimitive.vertex(x, y, 0, 0, drawRGB, drawAlpha);
	}
	
	public static void draw_vertex_color(double x, double y, int rgb, double alpha)
	{
		drawPrimitive.vertex(x, y, 0, 0, rgb, alpha);
	}
	
	public static void draw_vertex_texture(double x, double y, double tx, double ty)
	{
		drawPrimitive.vertex(x, y, tx, ty, drawRGB, drawAlpha);
	}
	
	public static void draw_vertex_texture_color(double x, double y, double tx, double ty, int rgb, double alpha)
	{
		drawPrimitive.vertex(x, y, tx, ty, rgb, alpha);
	}
	
	public static void draw_primitive_end()
	{
		if(drawPrimitiveTexture == null)
			EN.gl.glDisable(GL2ES1.GL_TEXTURE_2D);
		else
		{
			EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
			EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, drawPrimitiveTexture.getHandle());
		}
		drawPrimitive.render();
		drawPrimitiveTexture = null;
	}
    
    public static void draw_text(double x, double y, CharSequence str)
    {
    	draw_text_transformed(x, y, str, 1, 1, 0);
    }
    public static void draw_text_ext(double x, double y, CharSequence str, int sep, int w)
    {
    	draw_text_ext_transformed(x, y, str, sep, w, 1, 1, 0);
    }
    
    public static void draw_text_transformed(double x, double y, CharSequence str, double xscale, double yscale, double angle)
    {
    	draw_text_ext_transformed(x, y, str, -1, -1, xscale, yscale, angle);
    }
    
    public static void draw_text_ext_transformed(double x, double y, CharSequence str, int sep, int w, double xscale, double yscale, double angle)
    {
    	if(str.length() == 0)
    		return;
    	EN.gl.glPushMatrix();
    	EN.gl.glTranslatef((float)x, (float)y, 0);
    	EN.gl.glRotatef((float) angle, 0, 0, -1);
    	EN.gl.glScalef((float)xscale, (float)yscale, 1);
    	TextImpl text = textCache.get(font);
    	if(text == null) {
    		text = new TextImpl(EN.gl, font, Math.max(32, str.length()), BufferUsage.DYNAMIC_DRAW);
    		textCache.put(font, text);
    	}
    	font.build(EN.gl, text, str, sep, w, halign/2.0f, valign/2.0f);
    	applyDrawColor();
    	text.render(EN.gl);
    	EN.gl.glPopMatrix();
    }

    
	//////////////////////////////////////////////////////////////////////////
	// 3D Drawing
	//////////////////////////////////////////////////////////////////////////

	public static void d3d_start()
	{
		d3d_set_hidden(true);
		d3d_set_perspective(true);
		EN.gl.glAlphaFunc(GL2ES1.GL_GREATER, 0.0f);
		EN.gl.glEnable(GL2ES1.GL_ALPHA_TEST);
		
	}
	public static void d3d_end()
	{
		d3d_set_hidden(false);
		d3d_set_perspective(false);
		d3d_set_hidden(false);
		EN.gl.glDisable(GL2ES1.GL_ALPHA_TEST);
		EN.gl.glDisable(GL2ES1.GL_FOG);
	}
	
	public static void d3d_set_hidden(boolean b)
	{
		if(b)
		{
			EN.gl.glDepthFunc(GL2ES1.GL_LEQUAL);
			EN.gl.glEnable(GL2ES1.GL_DEPTH_TEST);
		}
		else
			EN.gl.glDisable(GL2ES1.GL_DEPTH_TEST);
	}
	
	public static void d3d_set_culling(boolean b)
	{
		if(b)
		{
			EN.gl.glCullFace(GL2ES1.GL_FRONT);
			EN.gl.glEnable(GL2ES1.GL_CULL_FACE);
		}
		else
			EN.gl.glDisable(GL2ES1.GL_CULL_FACE);
	}
	
	public static void d3d_set_perspective(boolean b)
	{
		perspective = b;
	}
	
	public static boolean d3d_get_perspective()
	{
		return perspective;
	}
    
    public static void d3d_set_fog(boolean enable, int color, double zNear, double zFar)
    {
    	if(enable)
    	{
    		EN.gl.glEnable(GL2ES1.GL_FOG);
            EN.gl.glFogf(GL2ES1.GL_FOG_MODE, GL2ES1.GL_LINEAR);
            EN.gl.glFogf(GL2ES1.GL_FOG_START, (float)zNear);
            EN.gl.glFogf(GL2ES1.GL_FOG_END, (float)zFar);
            ARGB.toRGBAf(view, 0, color|0xFF000000);
            EN.gl.glFogfv(GL2ES1.GL_FOG_COLOR, view, 0);
    	}
    	else
    		EN.gl.glDisable(GL2ES1.GL_FOG);
    }
    

	public static void d3d_set_projection_ortho(double x, double y, double w, double h, double angle)
	{
		EN.gl.glMatrixMode(GL2ES1.GL_PROJECTION);
		EN.gl.glLoadIdentity();
		EN.gl.glOrthof((float)x, (float)(x+w), (float)y, (float)(y+h), -65536.0f, 65535.0f);
		if(angle != 0)
			throw new UnsupportedOperationException("Ortho angle "+angle+" not implemented!");
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
		EN.gl.glLoadIdentity();
	}
	
	public static void d3d_set_projection(double xfrom, double yfrom, double zfrom, double xto, double yto, double zto, double xup, double yup, double zup)
	{
		EN.gl.glMatrixMode(GL2ES1.GL_PROJECTION);
		EN.gl.glLoadIdentity();
		int vi = Room.CURRENT.getViewIndex();
		int wport, hport;
		if(vi >= 0) {
			wport = Room.CURRENT.views[vi].wport;
			hport = Room.CURRENT.views[vi].hport;
		} else {
			wport = Room.CURRENT.info.width;
			hport = Room.CURRENT.info.height;
		}
		EN.glu.gluPerspective((45f*hport)/wport, (float)wport/hport, 1, 65536);
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
		EN.gl.glLoadIdentity();
		EN.gl.glScalef(1, -1, 1);
		EN.glu.gluLookAt((float)xfrom, (float)yfrom, (float)zfrom, (float)xto, (float)yto, (float)zto, xup, yup, zup);
	}
    
    public static void d3d_draw_wall(double x0, double y0, double z0, double x1, double y1, double z1, Tex2D texture, double hRepeat, double vRepeat)
    {
    	double d = Trig.dist(x0, y0, x1, y1);
    	if(d < 0.000001)
			return;
    	double nx = (y0-y1)/d, ny = (x1-x0)/d;
    	
    	d3d_primitive_begin_texture(PrimitiveType.TRIANGLE_STRIP, texture);
    	d3d_vertex_normal_texture(x0, y0, z0, nx, ny, 0, 0, 0);
    	d3d_vertex_normal_texture(x1, y1, z0, nx, ny, 0, hRepeat, 0);
    	d3d_vertex_normal_texture(x0, y0, z1, nx, ny, 0, 0, vRepeat);
    	d3d_vertex_normal_texture(x1, y1, z1, nx, ny, 0, hRepeat, vRepeat);
		d3d_primitive_end();
    }
    
    public static void d3d_draw_floor(double x0, double y0, double z0, double x1, double y1, double z1, Tex2D texture, double hRepeat, double vRepeat)
    {
    	double d = Trig.dist(x0, z0, x1, z1);
    	if(d < 0.000001)
			return;
    	double nx = (z1-z0)/d, nz = (x1-x0)/d;
    	
    	d3d_primitive_begin_texture(PrimitiveType.TRIANGLE_STRIP, texture);
    	d3d_vertex_normal_texture(x0, y0, z0, nx, 0, nz, 0, 0);
    	d3d_vertex_normal_texture(x1, y0, z1, nx, 0, nz, hRepeat, 0);
    	d3d_vertex_normal_texture(x0, y1, z0, nx, 0, nz, 0, vRepeat);
    	d3d_vertex_normal_texture(x1, y1, z1, nx, 0, nz, hRepeat, vRepeat);
		d3d_primitive_end();
    }
    
	public static void d3d_draw_block(double x0, double y0, double z0, double x1, double y1, double z1, Tex2D texture, double hRepeat, double vRepeat)
	{
		double xm = (x0+x1)/2;
		double ym = (y0+y1)/2;
		double zm = (z0+z1)/2;

		if(texture == null)
			EN.gl.glDisable(GL2ES1.GL_TEXTURE_2D);
		else
		{
			EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
			EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, texture.getHandle());
		}
		applyDrawColor();
		
		EN.gl.glPushMatrix();
		EN.gl.glTranslatef((float)xm, (float)ym, (float)zm);
		EN.gl.glScalef((float)(x1-xm), (float)(y1-ym), (float)(z1-zm));
		
		EN.gl.glMatrixMode(GL2ES1.GL_TEXTURE);
		EN.gl.glScalef((float)hRepeat, (float)vRepeat, 1);
		cube.render(EN.gl);

		EN.gl.glLoadIdentity();
		EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
		EN.gl.glPopMatrix();
	}

	
	public static void d3d_primitive_begin(PrimitiveType kind)
	{
		d3dPrimitiveTexture = null;
		d3dPrimitive.begin(kind);
	}
	
	public static void d3d_primitive_begin_texture(PrimitiveType kind, Tex2D texture)
	{
		d3dPrimitiveTexture = texture;
		d3dPrimitive.begin(kind);
	}
	
	public static void d3d_vertex(double x, double y, double z)
	{
		d3dPrimitive.vertex(x, y, z, 0, 0, 1, 0, 0, drawRGB, drawAlpha);
	}
	
	public static void d3d_vertex_normal(double x, double y, double z, double nx, double ny, double nz)
	{
		d3dPrimitive.vertex(x, y, z, nx, ny, nz, 0, 0, drawRGB, drawAlpha);
	}
	
	public static void d3d_vertex_color(double x, double y, double z, int rgb, double alpha)
	{
		d3dPrimitive.vertex(x, y, z, 0, 0, 1, 0, 0, rgb, alpha);
	}
	
	public static void d3d_vertex_normal_color(double x, double y, double z, double nx, double ny, double nz, int rgb, double alpha)
	{
		d3dPrimitive.vertex(x, y, z, nx, ny, nz, 0, 0, rgb, alpha);
	}
	
	public static void d3d_vertex_texture(double x, double y, double z, double tx, double ty)
	{
		d3dPrimitive.vertex(x, y, z, 0, 0, 1, tx, ty, drawRGB, drawAlpha);
	}
	
	public static void d3d_vertex_normal_texture(double x, double y, double z, double nx, double ny, double nz, double tx, double ty)
	{
		d3dPrimitive.vertex(x, y, z, nx, ny, nz, tx, ty, drawRGB, drawAlpha);
	}
	
	public static void d3d_vertex_texture_color(double x, double y, double z, double tx, double ty, int rgb, double alpha)
	{
		d3dPrimitive.vertex(x, y, z, 0, 0, 1, tx, ty, rgb, alpha);
	}
	
	public static void d3d_vertex_normal_texture_color(double x, double y, double z, double nx, double ny, double nz, double tx, double ty, int rgb, double alpha)
	{
		d3dPrimitive.vertex(x, y, z, nx, ny, nz, tx, ty, rgb, alpha);
	}
	
	public static void d3d_primitive_end()
	{
		if(d3dPrimitiveTexture == null)
			EN.gl.glDisable(GL2ES1.GL_TEXTURE_2D);
		else
		{
			EN.gl.glEnable(GL2ES1.GL_TEXTURE_2D);
			EN.gl.glBindTexture(GL2ES1.GL_TEXTURE_2D, d3dPrimitiveTexture.getHandle());
		}
		d3dPrimitive.render();
		d3dPrimitiveTexture = null;
	}
    
	//////////////////////////////////////////////////////////////////////////
	// Sound
	//////////////////////////////////////////////////////////////////////////
    
    public static void sound_global_volume(double val)
    {
    	Sound.globalVolume(Sound.linearize(val));
    }
    
    public static void sound_pan(String sound, double pan)
    {
    	Sound.get(sound).pan((float) pan);
    }
    
    public static void sound_play(String sound)
    {
    	Sound.get(sound).play();
    }
    
    public static void sound_loop(String sound)
    {
    	Sound.get(sound).loop();
    }
    public static void sound_stop(String sound)
    {
    	Sound.get(sound).stop();
    }
    public static boolean sound_isplaying(String sound)
    {
    	return Sound.get(sound).isplaying();
    }
    
    public static void sound_stop_all()
    {
    	Sound.stopAll();
    }

	//////////////////////////////////////////////////////////////////////////
	// User input
	//////////////////////////////////////////////////////////////////////////
    
    public static boolean keyboard_check_direct(VK key)
    {
    	return EN.ih.checkKeyDirect(key.newtKey);
    }
    
    public static boolean keyboard_check(VK key)
    {
    	return EN.in.isKeyDown(key.newtKey);
    }
    
    public static boolean keyboard_check_pressed(VK key)
    {
    	return EN.in.isKeyPressing(key.newtKey);
    }
    
    public static boolean keyboard_check_released(VK key)
    {
    	return EN.in.isKeyReleasing(key.newtKey);
    }
    
    private static int[] viewportSize = new int[2];
    private static float[] view = new float[4];
    
    ///The real thing would compute which view the mouse is in, unless rendering
    
	public static double mouse_x()
	{
		double mx = EN.in.getMouseX();
		double my = EN.in.getMouseY();
		
		Room.CURRENT.getViewportSize(viewportSize, 0);
		Letterbox.getViewport(view, 0, EN.getSurfaceWidth(), EN.getSurfaceHeight(), viewportSize[0], viewportSize[1]);
		mx = (mx-view[0])*viewportSize[0]/view[2];
		my = (my-view[1])*viewportSize[1]/view[3];
		
		int xp, wp, xv, wv;
		if(Room.CURRENT.info.viewsEnabled) {
			ViewData v = Room.CURRENT.views[Room.CURRENT.getViewIndex()];
			if(!Room.CURRENT.isRendering())
			{
				for(int i = Room.CURRENT.views.length-1; i>=0; i--)
				{
					ViewData vc = Room.CURRENT.views[i];
					if(vc.visible && mx>=vc.xport && my>=vc.yport && mx<vc.xport+vc.wport && my<vc.yport+vc.hport)
						v = vc;
				}
			}
			xp = v.xport; wp = v.wport; xv = v.xview; wv = v.wview;
		} else {
			xp = 0; wp = viewportSize[0]; xv = 0; wv = Room.CURRENT.info.width;
		}
		return xv + (mx-xp) * wv/wp;
	}

	public static double mouse_y()
	{
		double mx = EN.in.getMouseX();
		double my = EN.in.getMouseY();
		
		Room.CURRENT.getViewportSize(viewportSize, 0);
		Letterbox.getViewport(view, 0, EN.getSurfaceWidth(), EN.getSurfaceHeight(), viewportSize[0], viewportSize[1]);
		mx = (mx-view[0])*viewportSize[0]/view[2];
		my = (my-view[1])*viewportSize[1]/view[3];
		
		int yp, hp, yv, hv;
		if(Room.CURRENT.info.viewsEnabled) {
			ViewData v = Room.CURRENT.views[Room.CURRENT.getViewIndex()];
			if(!Room.CURRENT.isRendering())
			{
				for(int i = Room.CURRENT.views.length-1; i>=0; i--)
				{
					ViewData vc = Room.CURRENT.views[i];
					if(vc.visible && mx>=vc.xport && my>=vc.yport && mx<vc.xport+vc.wport && my<vc.yport+vc.hport)
						v = vc;
				}
			}
			yp = v.yport; hp = v.hport; yv = v.yview; hv = v.hview;
		} else {
			yp = 0; hp = viewportSize[1]; yv = 0; hv = Room.CURRENT.info.height;
		}
		return yv + (my-yp) * hv/hp;
	}
	
	public static boolean mouse_check_button_pressed(MB mb)
	{
		int index;
		switch (mb)
		{
		case LEFT:
			index = MouseEvent.BUTTON1;
			break;
		case MIDDLE:
			index = MouseEvent.BUTTON2;
			break;
		case RIGHT:
			index = MouseEvent.BUTTON3;
			break;
		default:
			return false;
		}
		return EN.in.isMousePressing(index);
	}
	
	//////////////////////////////////////////////////////////////////////////
	// File IO
	//////////////////////////////////////////////////////////////////////////

    public static class Ini implements AutoCloseable
    {
    	public FILE file;
    	public boolean modified;
    	public LinkedHashMap<String, LinkedHashMap<String, String>> table = new LinkedHashMap<String, LinkedHashMap<String,String>>();
    	
    	public static String key(String k)
    	{
    		return k.toLowerCase(Locale.ENGLISH);
    	}
    	
    	public double read_real(String s, String k, double def)
    	{
    		LinkedHashMap<String, String> section = table.get(key(s));
    		if(section == null)
    			return def;
    		String v = section.get(key(k));
    		if(v == null)
    			return def;
    		try {
    			return Double.parseDouble(v);
    		} catch(NumberFormatException err) {
    			return def;
    		}
    	}

		public<E> E read_enum(String s, String k, E[] values, E def)
		{
    		LinkedHashMap<String, String> section = table.get(key(s));
    		if(section == null)
    			return def;
    		String v = section.get(key(k));
    		if(v == null)
    			return def;
    		try {
    			int index = Integer.parseInt(v);
    			if(index < 0)
    				return def;
    			if(index >= values.length)
    				return def;
				return values[index];
    		} catch(NumberFormatException err) {
    			return def;
    		}
		}
		
		public<E> E read_enum(String s, String k, Map<Integer, E> values, E def)
		{
    		LinkedHashMap<String, String> section = table.get(key(s));
    		if(section == null)
    			return def;
    		String v = section.get(key(k));
    		if(v == null)
    			return def;
    		try {
    			int index = Integer.parseInt(v);
    			E value = values.get(index);
    			if(value == null)
    				return def;
    			return value;
    		} catch(NumberFormatException err) {
    			return def;
    		}
		}
		
		private static DigitalFormat FMT = new DigitalFormat("0.######");
		
		public void write_real(String s, String k, double val)
		{
			s = key(s);
    		LinkedHashMap<String, String> section = table.get(s);
    		if(section == null)
    		{
    			section = new LinkedHashMap<>();
    			table.put(s, section);
    		}
    		section.put(key(k), FMT.format(val));
			modified = true;
		}
    	
		@Override
		public void close() throws IOException
		{
			if(modified)
			{
				try(OutputStreamWriter fw = new OutputStreamWriter(file.write()))
				{
					for(Entry<String, LinkedHashMap<String, String>> section : table.entrySet())
					{
						if(section.getKey() != null) {
							fw.append('[');
							fw.append(section.getKey());
							fw.append(']');
							fw.append('\n');
						}
						for(Entry<String, String> kv : section.getValue().entrySet())
						{
							fw.append(kv.getKey());
							fw.append('=');
							fw.append(kv.getValue());
							fw.append('\n');
						}
					}
				}
				modified = false;
			}
		}
    }
    
    public static Ini ini_open(String fname) throws IOException
    {
    	Ini ini = new Ini();
    	ini.file = file(fname, true);
    	ini.file.getParentFile().mkdirs();
    	try(BufferedReader br = new BufferedReader(new InputStreamReader(ini.file.read())))
    	{
    		String sectionName = null;
    		LinkedHashMap<String, String> section = new LinkedHashMap<>();
    		ini.table.put(sectionName, section);
    		String in;
    		while((in = br.readLine()) != null)
    		{
    			in = in.trim();
    			if(in.startsWith(";"))
    				continue;
    			else if(in.startsWith("[") && in.endsWith("]")) {
    				sectionName = Ini.key(in.substring(1, in.length()-1));
    				section = new LinkedHashMap<>();
    				ini.table.put(sectionName, section);
    				
    			} else {
    				int sep = in.indexOf('=');
    				if(sep != -1)
    				{
    					String key = Ini.key(in.substring(0, sep).trim());
    					String value = in.substring(sep+1).trim();
    					section.put(key, value);
    				}
    			}
    		}
    	}
    	catch(FileNotFoundException e)
    	{
    		ini.modified = true;
    	}
    	return ini;
    }
    
    public static PrintWriter file_text_open_write(String fname) throws FileNotFoundException
    {
    	return new PrintWriter(file(fname, true).write());
    }
    
    public static BufferedReader file_text_open_read(String fname) throws FileNotFoundException
    {
    	return new BufferedReader(new InputStreamReader(file(fname, true).read()));
    }
    
    public static FILE file(String fname, boolean write)
    {
    	return new JAVAFILE(fname);
    }
    
	public static boolean file_exists(String fname, boolean write)
	{
		return file(fname, write).isFile();
	}
	
	public static boolean file_delete(String fname)
	{
		return file(fname, true).delete();
	}

	//////////////////////////////////////////////////////////////////////////
	// Rooms
	//////////////////////////////////////////////////////////////////////////

	public static String room()
	{
		return Room.CURRENT.info.name;
	}
	
	public static int room_width()
	{
		return Room.CURRENT.info.width;
	}
	
	public static int room_height()
	{
		return Room.CURRENT.info.height;
	}
	
	public static void transition_kind(int kind)
	{
		Room.transition = TransitionKind.VALUES[kind];
		Room.transitionSteps = 30;
	}
	
	public static void transition_steps(int steps)
	{
		Room.transitionSteps = 20;
	}
	
	public static void room_goto(String room) throws ActivityChangeException
	{
		throw new ActivityChangeException(ActivityChangeType.ROOM_CHANGE, RoomData.get(room));
	}
	
	public static void room_goto_next() throws ActivityChangeException
	{
		throw new ActivityChangeException(ActivityChangeType.ROOM_CHANGE, RoomData.list.get(Room.CURRENT.index+1));
	}
	
	public static void room_restart()
	{
		throw new ActivityChangeException(ActivityChangeType.ROOM_CHANGE, RoomData.list.get(Room.CURRENT.index));
	}
	
	public static void room_set_caption(String caption)
	{
		Room.CURRENT.setCaption(caption);
	}
	
	public static ViewData view()
	{
		return Room.CURRENT.views[Room.CURRENT.getViewIndex()];
	}
	
	//////////////////////////////////////////////////////////////////////////
	// Application
	///////////////////////////////////////////////

	public static boolean screen_save_part(String fname, int x0, int y0, int width, int height)
	{
		FILE f = file(fname, true);
		f.getParentFile().mkdirs();
		IImage image = EN.fbm.screenshot(x0, y0, width, height);
		try {
			PNGIO.save(f, image);
			return true;
		} catch(IOException err) {
			err.printStackTrace();
			return false;
		}
	}
	
	public static boolean screen_save(String fname)
	{
		return screen_save_part(fname, 0, 0, EN.fbm.getCurrentWidth(), EN.fbm.getCurrentHeight());
	}
	
	public static boolean game_save(String fname)
	{
		FILE f = file(fname, true);
		try (OutputStream out = f.write()) {	
			Room.save(out);
			return true;
		} catch(Exception err) {
			if(err instanceof ActivityChangeException)
				throw (ActivityChangeException)err;
			System.err.println("Failed to save: "+fname);
			err.printStackTrace();
			return false;
		}
	}
	
	public static boolean game_load(String fname)
	{
		FILE f = file(fname, true);
		try (InputStream in = f.read()) {
			Room.load(in);
			return true;
		} catch(Exception err) {
			if(err instanceof ActivityChangeException)
				throw (ActivityChangeException)err;
			System.err.println("Failed to load: "+fname);
			err.printStackTrace();
			return false;
		}
	}
	
	public static void game_end()
	{
		throw new ActivityChangeException(ActivityChangeType.GAME_END, (Activity)null);
	}
	
	public static void game_restart()
	{
		throw new ActivityChangeException(ActivityChangeType.GAME_RESTART, new Room(RoomData.list.get(0))); //Don't use persistent instance
	}
	
	public static double fps()
	{
		return EN.getFPS();
	}
	
	public static void window_set_fullscreen(boolean b)
	{
		EN.setFullscreen(b);
	}
}
