package owg.lgm;

import owg.lgm.gl.TexFormat;

public class IImage
{
	public enum Type
	{
		RGBA(4, TexFormat.RGBA), 
		RGB(3, TexFormat.RGB),
		LA(2, TexFormat.LUMINANCE_ALPHA), 
		GRAY(1, TexFormat.LUMINANCE),
		ALPHA(1, TexFormat.ALPHA);

		public final int bytesPerComponent;
		public final TexFormat texFormat;

		private Type(int bytes, TexFormat texFormat)
		{
			this.bytesPerComponent = bytes;
			this.texFormat = texFormat;
		}
	}
	public final String name;
	public final byte[] data;
	public final int width, height;
	public final Type type;

	public IImage(Type type, byte[] data, int width, int height, String name)
	{
		super();
		this.name = name;
		this.data = data;
		this.width = width;
		this.height = height;
		this.type = type;
	}

	public IImage(Type type, int width, int height, String name)
	{
		super();
		this.name = name;
		this.data = new byte[width*height*type.bytesPerComponent];
		this.width = width;
		this.height = height;
		this.type = type;
	}

	public IImage toRGBA()
	{
		if(type == Type.RGBA)
			return this;
		IImage r = new IImage(Type.RGBA, width, height, name);
		if(type == Type.GRAY)
		{
			for(int i = 0, j = 0; i<data.length; i++, j += 4)
			{
				r.data[j  ] = data[i];
				r.data[j+1] = data[i];
				r.data[j+2] = data[i];
				r.data[j+3] = (byte) 0xFF;
			}
		} 
		else if(type == Type.LA) 
		{
			for(int i = 0, j = 0; i<data.length; i += 2, j += 4)
			{
				r.data[j  ] = data[i];
				r.data[j+1] = data[i];
				r.data[j+2] = data[i];
				r.data[j+3] = data[i+1];
			}
		}
		else
		{
			for(int i = 0, j = 0; i<data.length; i += 3, j += 4)
			{
				r.data[j  ] = data[i  ];
				r.data[j+1] = data[i+1];
				r.data[j+2] = data[i+2];
				r.data[j+3] = (byte) 0xFF;
			}
		}
		return r;
	}
	
	public IImage toAlpha()
	{
		byte[] newData;
		if(type == Type.GRAY)
			newData = data;
		else
		{
			newData = new byte[width*height];
			for(int i = type.bytesPerComponent-1, j = 0; j<newData.length; i += type.bytesPerComponent, j++)
				newData[j] = data[i];
		}
		return new IImage(Type.ALPHA, newData, width, height, name);
	}
	
	public void flip() {
		byte[] copyBuffer = new byte[width*4];
		for(int y = 0; y<height/2; y++)
		{
			int a = y*width*4;
			int b = (height-1-y)*width*4;
			System.arraycopy(data, a, copyBuffer, 0, width*4);
			System.arraycopy(data, b, data, a, width*4);
			System.arraycopy(copyBuffer, 0, data, b, width*4);
		}
	}
}
