package owg.lgm.ar1.accessories;

import static owg.lgm.plaf.Engine.EN;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.nio.charset.StandardCharsets;
import java.util.ArrayList;
import java.util.Arrays;

import com.jogamp.opengl.GL2ES1;

import owg.lgm.Activity;
import owg.lgm.ResourceException;
import owg.lgm.gl.BMFontRenderer;
import owg.lgm.gl.BufferUsage;
import owg.lgm.gl.GradientText;
import owg.lgm.gl.Text;
import owg.lgm.gl.TextImpl;
import owg.lgm.in.EventQueue;
import owg.lgm.in.MB;
import owg.lgm.in.MouseEventType;
import owg.lgm.in.VK;
import owg.lgm.room.ActivityChangeException;
import owg.lgm.room.ActivityChangeType;
import owg.lgm.room.Room;
import owg.lgm.util.FILE;
import owg.lgm.util.Letterbox;
import owg.util.math.Clamp;
import owg.util.math.Vec1;

import static owg.lgm.GMLib.*;

import static owg.lgm.ar1.accessories.AR1Globals.global;

public class AR1Help implements Activity
{
	protected static class Line {
		int style = 0;
		String text;
	}
	
	protected final ArrayList<Line> lines = new ArrayList<>();
	protected final int[] counts = new int[4];
	
	protected final transient BMFontRenderer[] fonts = {AR1Fonts.A_12, AR1Fonts.A_26, AR1Fonts.A_16, AR1Fonts.A_14};
	protected final transient Text[] text = new Text[4];
	protected final int[] heights = {19, 35, 23, 21};
	
	protected double vsp = 0;
	
	protected int[] accum = new int[3];
	
	protected boolean mouseDrag = false;
	protected double dragStart;
	protected double dragOffset;
	
	public AR1Help() {}

	@Override
	public void start(GL2ES1 gl, ActivityChangeType type)
	{
		text[0] = new TextImpl(gl, fonts[0], 128, BufferUsage.DYNAMIC_DRAW);
		for(int i = 1; i<4; i++)
			text[i] = new GradientText(gl, fonts[i], 128, c_red, c_yellow, BufferUsage.DYNAMIC_DRAW);
		
		FILE f = file("res/help.txt", false);
		try(	InputStream in = f.read();
				InputStreamReader sr = new InputStreamReader(in, StandardCharsets.UTF_8);
				BufferedReader br = new BufferedReader(sr)) {
			String str;
			while((str=br.readLine()) != null)
			{
				Line line = new Line();
				while(line.style < str.length() && str.charAt(line.style) == '#' && line.style < 3)
					line.style++;
				counts[line.style]++;
				line.text = str.substring(line.style);
				lines.add(line);
			}
		} catch(IOException err) {
			throw new ResourceException("Failed to read "+f, err);
		}
	}

	@Override
	public void run(GL2ES1 gl)
	{	
		step();
		render();
	}

	protected int[] viewportSize = new int[2];
	protected float[] view = new float[4];

	protected void step()
	{
		double mx = EN.in.getMouseX();
		double my = EN.in.getMouseY();

		Room.CURRENT.getViewportSize(viewportSize, 0);
		Letterbox.getViewport(view, 0, EN.getSurfaceWidth(), EN.getSurfaceHeight(), viewportSize[0], viewportSize[1]);
		mx = (mx-view[0])/view[2];
		my = (my-view[1])/view[3];

		mx *= 640;
		my *= 480;

		int totalHeight = 0;
		for(int i = 0; i<4; i++)
			totalHeight += counts[i] * heights[i];
		
		EventQueue<MouseEventType, MB> mouseEvents = EN.in.getMouseEvents();
		for(int i = 0; i<mouseEvents.size(); i++)
		{
			if(mouseEvents.ids.get(i) == MB.LEFT)
			{
				if(mouseEvents.types.get(i) == MouseEventType.PRESS && mouseDrag == false)
				{
					mouseDrag = true;
					dragStart = my;
					dragOffset = global.helpScroll;
					vsp = 0;
				}
				else if(mouseEvents.types.get(i) == MouseEventType.RELEASE)
				{
					vsp = (accum[0]+accum[1]+accum[2])/3.0;
					mouseDrag = false;
					Arrays.fill(accum, 0);
				}
			}
		}
		
		if(mouseDrag)
		{
			int ty = Clamp.clampH((int) (dragOffset - my + dragStart), 0, totalHeight-480);
			accum[2] = accum[1];
			accum[1] = accum[0];
			accum[0] = global.helpScroll-ty;
			global.helpScroll = ty;
		}
		
		int dy = 0;
		if(keyboard_check(VK.UP))
			dy++;
		if(keyboard_check(VK.DOWN))
			dy--;
		if(dy != 0 && !mouseDrag)
		{
			//if(vsp*dy <= 0)
			//	vsp = dy*4;
			//else
			vsp += dy;
		}
		else
			vsp = Vec1.accelerate(vsp, -1);
		
		int tmp = global.helpScroll-(int)vsp;
		global.helpScroll = Clamp.clampH(tmp, 0, totalHeight-480);
		if(global.helpScroll != tmp)
			vsp = 0;
		
		if(keyboard_check_pressed(VK.ENTER) || keyboard_check_pressed(VK.ESCAPE))
			throw new ActivityChangeException(ActivityChangeType.PAUSE_RESUME, Room.CURRENT);
	}

	protected void render()
	{
		if(EN.fbm.beginRender(EN.getSurfaceWidth(), EN.getSurfaceHeight())) {
			EN.gl.glViewport(0, 0, 640, 480);
			EN.gl.glMatrixMode(GL2ES1.GL_PROJECTION);
			EN.gl.glLoadIdentity();
			EN.gl.glOrthof(0, 640, 0, 480, -32768, 32767);
			EN.gl.glMatrixMode(GL2ES1.GL_MODELVIEW);
			EN.gl.glLoadIdentity();

			renderDialog();
			EN.fbm.endRender();
		}
		EN.clear();
		EN.fbm.blit(EN.getSurfaceWidth(), EN.getSurfaceHeight());
	}

	protected void renderDialog() 
	{
		EN.gl.glClearColor(0x3e/255f, 0x40/255f, 0x86/255f, 1.0f);
		EN.gl.glClear(GL2ES1.GL_COLOR_BUFFER_BIT);
		EN.gl.glColor4f(0, 1, 0, 1);
		
		EN.gl.glPushMatrix();
		EN.gl.glTranslatef(4, -global.helpScroll, 0);
		
		int y = -global.helpScroll;
		for(int i = 0; i<lines.size(); i++)
		{
			Line line = lines.get(i);
			if(y < 480 && y > -heights[line.style])
			{
				fonts[line.style].build(EN.gl, text[line.style], line.text, -1, -1, 0, 0);
				text[line.style].render(EN.gl);
			}
			EN.gl.glTranslatef(0, heights[line.style], 0);
			y += heights[line.style];
		}
		EN.gl.glPopMatrix();
	}
	
	@Override
	public void end(GL2ES1 gl, ActivityChangeType type)
	{
		for(int i = 0; i<4; i++)
			text[i].dispose();
	}

	public void start()
	{
		throw new ActivityChangeException(ActivityChangeType.PAUSE_RESUME, this);
	}
}
