package owg.lgm.ar1.accessories;

import static owg.lgm.GMLib.*;
import static owg.lgm.ar1.accessories.AR1Globals.global;
import static owg.lgm.plaf.Engine.EN;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;

import owg.lgm.Sprite;
import owg.lgm.ar1.Animal;
import owg.lgm.midi.MidiThread;
import owg.lgm.room.Room;
import owg.lgm.room.Room.ViewData;
import owg.lgm.util.Letterbox;
import owg.util.rethrow.Rethrow;

public class AR1Scripts 
{
	public static MidiThread midi;
	public static void initMidi()
	{
		if(midi == null)
		{
			midi = new MidiThread(new File(EN.dir, "res/Midi"), new File(EN.dir, "res/A320U.sf2"));
			midi.start();
		}
	}
	
	public static byte[] tempSav;
	public static void saveTemp()
	{
		try(ByteArrayOutputStream baos = new ByteArrayOutputStream(192000)) {
			Room.save(baos);
			tempSav = baos.toByteArray();
		} catch (IOException e) {
			Rethrow.rethrow(e);
		}
	}
	public static void loadTemp()
	{
		try(ByteArrayInputStream bain = new ByteArrayInputStream(tempSav)) {
			Room.load(bain);
		} catch (IOException e) {
			Rethrow.rethrow(e);
		}
	}
	
	public static void showMainMenu()
	{
		new AR1Dialog()
		{
			@Override
			public void dialog() throws ConverseException, SleepException
			{
				int q=show_message_ext("Welcome to Oddwarg Animal RPG","New Game","Load Game", "Options","Quit Game");
				if(q == 0)
					showNewGame(() -> showMainMenu());
				else if(q == 1)
					showLoadGame(() -> showMainMenu());
				else if(q == 2)
					showOptions(() -> showMainMenu());
				else
					game_end();
			}
		}.start();
	}
	
	public static String saveName(int no, boolean image)
	{
		return "AR1_save"+no+(image?".png":".sav");
	}
	
	public static String saveName(boolean image)
	{
		return saveName(global.currentFile, image);
	}
	
	public static void save()
	{
		Animal animal = instance(Animal.class);
		if(animal != null)
		{
			game_save(saveName(false));
			ViewData view = Room.CURRENT.views[0];
			
			int x0 = (int) (animal.x-80-view.xview);
			int y0 = (int) (animal.y-100-view.yview);
			if(x0 < 0) x0 = 0;
			if(y0 < 0) y0 = 0;
			if(x0 > 640-160)
				x0 = 640-160;
			if(y0 > 480-160)
				y0 = 480-160;
			screen_save_part(saveName(true), (int)x0, (int)y0, 160, 160);
		}
	}
	
	public static void showNewGame(Runnable fail)
	{
		new AR1Dialog()
		{
			int drawChar = -2;
			@Override
			public void dialog() throws ConverseException, SleepException
			{
				global.cha = 0;
				global.name = "Nobody";
				global.currentFile=show_menu("Choose a save slot","Cancel","File1","File2","File3");
				if(global.currentFile == 0)
				{
					fail.run();
					return;
				}
				if(file_exists(saveName(false), true))
				{
					if(!show_question("That save already exists. Do you want to overwrite it?"))
					{
						fail.run();
						return;
					}
				}
				drawChar = -1;
				global.cha=show_message_ext("Choose your looks","<Male","^Kaseta",">Female");
				drawChar = global.cha;
				global.name=get_string("Type in your name", "Nobody");
				instance(Animal.class).start();
			}

			Sprite ystandr = Sprite.get("ystandr");
			Sprite ksitr = Sprite.get("ksitr");
			Sprite flier = Sprite.get("flier");

			@Override
			protected void renderDialog()
			{
				EN.clear();
				if(drawChar == -1 || drawChar == 0)
					draw_sprite(ystandr,0,128,128);
				if(drawChar == -1 || drawChar == 1)
					draw_sprite(ksitr,0,320,128);
				if(drawChar == -1 || drawChar == 2)
					draw_sprite(flier,0,640-128,128);
				super.renderDialog();
			}
		}.start();
	}
	
	public static void showLoadGame(Runnable fail)
	{
		new AR1Dialog()
		{	
            Sprite sprite321 = Sprite.get("sprite321");
            Sprite sprite322 = Sprite.get("sprite322");
            Sprite sprite323 = Sprite.get("sprite323");
            Sprite nosave = Sprite.get("nosave");
            
            Sprite game0, game1, game2;
            Boolean loadResult = null;
            
            @Override
			public void start(com.jogamp.opengl.GL2ES1 gl, owg.lgm.room.ActivityChangeType type) {
            	super.start(gl, type);
            	
                if (file_exists(saveName(1, false), true)) {
                	game0 = sprite321;
                	sprite321.replace(EN.gl, file(saveName(1, true), true), 1, false, false, 128, 96);
                } else
                	game0 = nosave;
                if (file_exists(saveName(2, false), true)) {
                	game1 = sprite322;
                	sprite322.replace(EN.gl, file(saveName(2, true), true), 1, false, false, 128, 96);
                } else
                	game1 = nosave;
                if (file_exists(saveName(3, false), true)) {
                	game2 = sprite323;
                	sprite323.replace(EN.gl, file(saveName(3, true), true), 1, false, false, 128, 96);
	            } else
	            	game2 = nosave;
            }
            
			@Override
			public void dialog() throws ConverseException
			{
                int qwern=show_message_ext("Which file to LOAD?","None","File1","File2","File3");
                if(qwern == 0)
                {
                	fail.run();
                	return;
                }
                if(loadResult == null)
                	loadResult = game_load(saveName(qwern, false)); 
                if(!loadResult)
                {
                	show_message("Cannot load that file!");
                	fail.run();
                	return;
                }
			}
			@Override
			protected void renderDialog() 
			{
                draw_set_color(c_black);
                draw_rectangle(0, 0, 640, 480, false);
                if(game0 != null)
                	draw_sprite(game0,0,640-80*5,96);
                if(game1 != null)
                	draw_sprite(game1,0,640-80*3,96);
                if(game2 != null)
                	draw_sprite(game2,0,640-80,96);
				super.renderDialog();
			}
		}.start();
	}
	

	public static void showOptions(Runnable fail)
	{
		new AR1Dialog()
		{	            
			@Override
			public void dialog() throws ConverseException
			{
                int r=show_menu("Options", 
                		"Fullscreen: "+(EN.isFullscreen()?"Yes":"No"), 
                		"Integer Scaling: "+("true".equals(System.getProperty(Letterbox.INTEGER_SCALING))?"Yes":"No"), 
                		"Vertical Sync: "+(EN.getVSync()?"Yes":"No"), 
                		"Go back");
                if(r==0)
                	EN.setFullscreen(!EN.isFullscreen());
                else if(r==1)
                	System.setProperty(Letterbox.INTEGER_SCALING, String.valueOf(!("true".equals(System.getProperty(Letterbox.INTEGER_SCALING)))));
                else if(r==2)
                	EN.setVSync(!EN.getVSync());
                else
                {
                	fail.run();
                	return;
                }
                showOptions(fail);
			}
		}.start();
	}
	
	public static void showGameOver()
	{
		new AR1Dialog()
		{
			@Override
			public void dialog() throws ConverseException
			{
	            int qhu=show_message_ext("You are dead","Load","Main Menu","Quit");
	            if (qhu==0) {
	            	if(!game_load(saveName(false)))
	            	{
	            		show_message("Failed to load the file!");
	            		showGameOver();
	            	}
	            } else if (qhu==1) {
	            	sound_stop_all();
	            	midi.midiStop();
	            	game_restart();
	            } else if (qhu==2)
	            	game_end();
			}
		}.start();
	}
	
	public static void showMap()
	{
		saveTemp();
		transition_kind(0);
		room_goto("map");
	}
	
	public static void showItems()
	{
		saveTemp();
		transition_kind(0);
		room_goto("Allitems95");
	}
	
	public static void showHelp()
	{
		new AR1Help().start();
	}
	
	public static void showUseItem()
	{
		new AR1Dialog()
		{
			@Override
			public void dialog() throws ConverseException, SleepException
			{
				int itemm=show_menu("What to use?",
						"Nothing",
						"Potion ("+(global.potion[0])+")",
						"Full Health ("+(global.potion[1])+")",
						"Increase MaxHealth ("+(global.potion[2])+")",
						"Extra Strength ("+(global.potion[3])+")",
						"Cure Sickness ("+(global.potion[4])+")",
						"Poison ("+(global.potion[5])+")");
				if (itemm>0)
				{
					itemm += -1;
					if (show_question("Use " + (global.potionname[itemm])+"?"))
					{
						if (global.potion[itemm]>0)
						{
							show_message("You used " + (global.potionname[itemm])+".");
							global.potion[itemm] += -1;
							if (itemm==0)
								global.life += 32;
							if (itemm==1)
								global.life = global.maxlife;
							if (global.life>global.maxlife)
								global.life = global.maxlife;
							if (itemm==2)
							{
								double qk = global.maxlifelev/4.0;
								global.maxlife += 32/qk;
								global.maxlifelev += 1;
							}
							if (itemm==3)
							{
								double mk = global.sharpreal/4.0;
								global.sharpness += 1.0/mk + Math.pow(mk-1, 1.5)/20;
								global.sharpreal += 1;
							}
							if (itemm==4)
								global.poisoned = 0;
							if (itemm==5)
								global.poisoned = 1;
						}
						else
							show_message("You dont have that item.");
					}
				}
			}
		}.start();
	}
}
