package owg.lgm.gl;

import java.lang.ref.WeakReference;

import com.jogamp.opengl.GL2ES1;

import owg.lgm.util.DestructibleImpl;
import owg.lgm.util.FixedIndexList;
import owg.util.autoname.AutoName;

public abstract class GLResource extends DestructibleImpl
{
	protected static final FixedIndexList<WeakReference<GLResource>> registry = new FixedIndexList<>();
	
	public static void initAll(GL2ES1 gl)
	{
		for(int pass = 0; pass<2; pass++)
		{
			for(int i = 0; i<registry.size(); i++)
			{
				if(registry.containsIndex(i))
				{
					WeakReference<GLResource> ref = registry.get(i);
					GLResource res = ref.get();
					if(res != null)
					{
						if(res.getInitOrder() == pass)
							res.init(gl);
					}
					else
					{
						registry.removeIndex(i);
						i--;
					}
				}
			}
		}
	}

	public static void uninitAll(GL2ES1 gl)
	{
		for(int i = 0; i<registry.size(); i++)
		{
			if(!registry.containsIndex(i))
				continue;
			WeakReference<GLResource> ref = registry.get(i);
			if(ref != null)
			{
				GLResource res = ref.get();
				if(res != null)
					res.uninit();
				else
				{
					registry.removeIndex(i);
					i--;
				}
			}
		}
	}
	
	protected String name;
	protected transient final int regIndex;
	
	public GLResource(String name)
	{
		this.name = name;
		regIndex = registry.add(new WeakReference<GLResource>(this));
	}
	
	public GLResource()
	{
		this.name = AutoName.create(getClass());
		regIndex = registry.add(new WeakReference<GLResource>(this));
	}
	
	@Override
	public void dispose()
	{
		if(!isDisposed())
		{
			uninit();
			registry.removeIndex(regIndex);
		}
		super.dispose();
	}
	
	public String getName()
	{
		return name;
	}
	public void setName(String name)
	{
		this.name = name;
	}
	
	@Override
	public String toString()
	{
		return getClass().getSimpleName()+": "+name;
	}
	
	public abstract int getInitOrder();
	
	public abstract void init(GL2ES1 gl);
	
	public abstract void uninit();
	
	@Override
	protected void finalize()
	{
		if(!isDisposed())
			System.err.println("Leaked "+toString());
	}
}
