package owg.lgm.gl;

import com.jogamp.opengl.GL2ES1;
import com.jogamp.opengl.GL3;

import owg.lgm.util.AutoDestructibleImpl;

public class Mesh extends AutoDestructibleImpl
{
	public static final Mesh EMPTY = new Mesh() {
		@Override
		public void render(GL2ES1 gl)
		{
			//No-op
		}
	};
	
	public final AttribArray[] vertexAttributes;
	public DrawCommand drawCommand;
	
	public Mesh(DrawCommand drawCommand, AttribArray... vertexAttributes)
	{
		this.drawCommand = drawCommand;
		this.vertexAttributes = vertexAttributes;
	}
	
	protected Mesh()
	{
		vertexAttributes = null;
	}
	
	public void render(GL2ES1 gl)
	{
		int vbo = -1;
		for (int i = 0; i < vertexAttributes.length; i++)
		{
			AttribArray array = vertexAttributes[i];
			if(vbo != array.vertexBufferObject.handle)
			{
				gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, array.vertexBufferObject.handle);
				vbo = array.vertexBufferObject.handle;
			}
			array.attrib.enable(gl, array);
		}
		drawCommand.invoke(gl);
		for (int i = vertexAttributes.length-1; i >= 0; i--)
		{
			AttribArray array = vertexAttributes[i];
			array.attrib.disable(gl, array);
		}
		//gl.glBindBuffer(GL3.GL_ARRAY_BUFFER, 0);
	}
}
