package owg.lgm.plaf;

import static owg.lgm.plaf.Engine.EN;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import com.jogamp.newt.event.InputEvent;
import com.jogamp.newt.event.KeyEvent;
import com.jogamp.newt.event.KeyListener;
import com.jogamp.newt.event.MouseAdapter;
import com.jogamp.newt.event.MouseEvent;
import com.jogamp.newt.event.WindowEvent;
import com.jogamp.newt.event.WindowListener;
import com.jogamp.newt.event.WindowUpdateEvent;
import com.jogamp.newt.opengl.GLWindow;

import jogamp.opengl.util.pngj.ImageInfo;
import jogamp.opengl.util.pngj.PngWriter;
import owg.lgm.IImage;
import owg.lgm.IImage.Type;
import owg.lgm.in.InputState;
import owg.lgm.in.VK;
import owg.util.exceptionless.Sleep;

public class InputHandler extends MouseAdapter implements KeyListener, WindowListener
{	
	/**The lock to be claimed by threads that need to change the IO state.*/
	protected final Object swapLock = new Object();
	/**The current input state. Must not be modified. Is swapped once every step to reflect new values.*/
	protected InputState currentIOState = new InputState();
	/**The current input state. May be modified provided the swapLock is claimed.*/
	protected InputState nextIOState = new InputState();
	
	protected boolean gamepadMode = false;
	protected boolean reloaded = false;
	
	public void init(GLWindow target)
	{
		nextIOState.setFocused(target.hasFocus());
		nextIOState.setSurfaceSize(target.getSurfaceWidth(), target.getSurfaceHeight());
		
		target.addKeyListener(this);
		target.addMouseListener(this);
		target.addWindowListener(this);
	}
	
	protected VK waitKey;

	public void updateModifiers(InputEvent e)
	{
		synchronized(swapLock)
		{
			int modifiers = e.getModifiers();
			boolean ctrl = (modifiers&InputEvent.CTRL_MASK) != 0;
			if(ctrl && !nextIOState.isKeyDown(KeyEvent.VK_CONTROL))
				nextIOState.pressKey(KeyEvent.VK_CONTROL, '\0', modifiers, false);
			else if(!ctrl && nextIOState.isKeyDown(KeyEvent.VK_CONTROL))
				nextIOState.releaseKey(KeyEvent.VK_CONTROL);
			
			boolean alt = (modifiers&InputEvent.ALT_MASK) != 0;
			if(alt && !nextIOState.isKeyDown(KeyEvent.VK_ALT))
				nextIOState.pressKey(KeyEvent.VK_ALT, '\0', modifiers, false);
			else if(!alt && nextIOState.isKeyDown(KeyEvent.VK_ALT))
				nextIOState.releaseKey(KeyEvent.VK_ALT);
			
			
			boolean shift = (modifiers&InputEvent.SHIFT_MASK) != 0;
			if(shift && !nextIOState.isKeyDown(KeyEvent.VK_SHIFT))
				nextIOState.pressKey(KeyEvent.VK_SHIFT, '\0', modifiers, false);
			else if(!shift && nextIOState.isKeyDown(KeyEvent.VK_SHIFT))
				nextIOState.releaseKey(KeyEvent.VK_SHIFT);
		}
	}
	
	@Override
	public void windowDestroyNotify(WindowEvent e) {
		if(Engine.EN.gameConfig.treatCloseAsEscape)
		{
			synchronized (swapLock)
			{
				nextIOState.pressKey(KeyEvent.VK_ESCAPE, '\0', nextIOState.getModifiers(), false);
				nextIOState.releaseKey(KeyEvent.VK_ESCAPE);
			}
		}
		else
			Engine.EN.close();
	}
	@Override
	public void windowGainedFocus(WindowEvent e) {
		synchronized (swapLock)
		{
			nextIOState.setFocused(true);
		}
	}
	@Override
	public void windowLostFocus(WindowEvent e) {
		synchronized (swapLock)
		{
			nextIOState.setFocused(false);
		}
	}
	
	@Override
	public void windowDestroyed(WindowEvent e) {}
	@Override
	public void windowMoved(WindowEvent e) {}
	@Override
	public void windowRepaint(WindowUpdateEvent e) {}
	@Override
	public void windowResized(WindowEvent e) {
		synchronized (swapLock)
		{
			nextIOState.setSurfaceSize(Engine.EN.window.getSurfaceWidth(), Engine.EN.window.getSurfaceHeight());
		}
	}
	
	@Override
	public void keyReleased(KeyEvent e)
	{
		if((e.getModifiers()&InputEvent.AUTOREPEAT_MASK) == 0)
		{
			synchronized (swapLock)
			{
				updateModifiers(e);
				nextIOState.releaseKey(e.getKeySymbol());
			}
		}
	}
	
	@Override
	public void keyPressed(KeyEvent e)
	{
		if(e.isControlDown() && e.getKeyCode() == KeyEvent.VK_ENTER) {
			System.out.println("Animator restart");
			EN.animator.start();
		}
		else if(e.getModifiers() == 0 && e.getKeyCode() == KeyEvent.VK_F4)
		{
			if(EN.gameConfig.letF4SwitchFullscreen)
				EN.animator.invokeOnAnimator(() -> EN.setFullscreen(!EN.isFullscreen()));
		}
		else if(e.getModifiers() == 0 && e.getKeyCode() == KeyEvent.VK_F9)
		{
			if(EN.gameConfig.letF9Screenshot)
			{
				EN.window.invoke(false, (gl) -> {
					File f;
					int i = 0;
					do {
						f = new File(EN.dir, "screenshot"+i+".png");
						i++;
					} while(f.exists());
					
					try(FileOutputStream os = new FileOutputStream(f)) {
						IImage scr = EN.fbm.screenshot();
						ImageInfo inf = new ImageInfo(scr.width, scr.height, 8, scr.type == Type.RGBA);
						PngWriter pngw = new PngWriter(os, inf, f.getName());
						byte[] row = new byte[scr.width*scr.type.bytesPerComponent];
						for(int y = 0; y<scr.height; y++)
						{
							System.arraycopy(scr.data, y*row.length, row, 0, row.length);
							pngw.writeRowByte(row, y);
						}
						pngw.end();
					} catch(IOException err) {
						err.printStackTrace();
					}
					return true;
				});
			}
		}
		
		synchronized (swapLock)
		{
			if(waitKey != null)
			{
				if(waitKey == VK.ANY_KEY || waitKey.newtKey == e.getKeySymbol())
					waitKey = null;
			}
			gamepadMode = false;
			nextIOState.pressKey(e.getKeySymbol(), e.getKeyChar(),
				e.getModifiers(), (e.getModifiers()&InputEvent.AUTOREPEAT_MASK) != 0);	
			updateModifiers(e);				
		}
	}
	
	@Override
	public void mouseDragged(MouseEvent e)
	{
		updateModifiers(e);
		synchronized (swapLock)
		{
			nextIOState.dragMouse(e.getX(), e.getY(), e.getButton(), e.getModifiers());
		}
	}
	@Override
	public void mouseMoved(MouseEvent e)
	{
		updateModifiers(e);
		synchronized (swapLock)
		{
			nextIOState.moveMouse(e.getX(), e.getY(), e.getModifiers());
		}
	}
	@Override
	public void mousePressed(MouseEvent e)
	{
		synchronized (swapLock)
		{
			updateModifiers(e);
			nextIOState.pressMouseButton(e.getX(), e.getY(), e.getButton(), e.getModifiers());
		}
	}
	@Override
	public void mouseEntered(MouseEvent e)
	{
		synchronized (swapLock)
		{
			updateModifiers(e);
			nextIOState.mouseEntered(e.getX(), e.getY(), e.getModifiers());					
		}
	}
	@Override
	public void mouseExited(MouseEvent e)
	{
		synchronized (swapLock)
		{
			updateModifiers(e);
			nextIOState.mouseExited(e.getX(), e.getY(), e.getModifiers());					
		}
	}
	@Override
	public void mouseReleased(MouseEvent e)
	{
		synchronized (swapLock)
		{
			updateModifiers(e);
			nextIOState.releaseMouseButton(e.getX(), e.getY(), e.getButton(), e.getModifiers());					
		}
	}
	@Override
	public void mouseWheelMoved(MouseEvent e)
	{
		synchronized (swapLock)
		{
			updateModifiers(e);
			nextIOState.mouseWheelMoved(e.getRotation()[0], -e.getRotation()[1], e.getModifiers());					
		}
	}
	
	public void clear()
	{
		synchronized (swapLock)
		{
			currentIOState.clear();
			nextIOState.clear();
		}
	}

	public InputState swap()
	{
		synchronized (swapLock)
		{
			nextIOState.transfer(currentIOState);
			InputState tmp = currentIOState;
			currentIOState = nextIOState;
			nextIOState = tmp;
			
			if(reloaded)
			{
				currentIOState.clear();
				nextIOState.clear();
				gamepadMode = false;
				reloaded = false;
			}
			
			return currentIOState;
		}
	}
	
	
	public void gamepadKeyPress(int vk)
	{
		gamepadMode = true;
		nextIOState.pressKey(vk, '\0', 0, false);
	}
	public void gamepadKeyRelease(int vk)
	{
		nextIOState.releaseKey(vk);
	}

	public boolean checkKeyDirect(short key) {
		synchronized (swapLock) {
			return nextIOState.isKeyDown(key);		
		}
	}

	public void waitUntilKey(int stepSize, int timeout, VK key) {
		synchronized (swapLock) {
			waitKey = key;			
		}
		long startTime = System.currentTimeMillis();
		while(true)
		{
			if(System.currentTimeMillis() - startTime >= timeout)
				return;
			synchronized (swapLock) {
				if(waitKey == null)
					return;
			}
			Sleep.sleep(stepSize);
		}
	}
	
	public String toString(VK key)
	{
		String nm = key.name();
		int sep = nm.indexOf('_');
		if(sep == 0)
			return nm.substring(1);
		else if(sep == -1)
		{
			return nm;
			//return nm.charAt(0)+nm.substring(1).toLowerCase(Locale.ENGLISH);
		}
		else
		{
			return nm.replace('_', ' ');
			//return nm.charAt(0)+nm.substring(1, sep).toLowerCase(Locale.ENGLISH)+' '+
			//		nm.charAt(sep+1)+nm.substring(sep+2).toLowerCase(Locale.ENGLISH);
		}
	}
}
