package owg.lgm.plaf;

import java.io.File;

import com.jogamp.common.util.locks.RecursiveLock;
import com.jogamp.newt.Display;
import com.jogamp.newt.opengl.GLWindow;

import owg.util.nativelibraryloader.NativeLibraryLoader;
import owg.util.nativelibraryloader.NativeLibraryLoader.PlatformSupport;
import owg.util.xkb.XkbDetectableAutoRepeat;

/**This class applies the {@link XkbDetectableAutoRepeat} fix if appropriate, when calling {@link #apply(Display, GLWindow)}.*/
public class XKBFix {
	protected static final PlatformSupport LINUX = new PlatformSupport(
			(osName, osArch) -> NativeLibraryLoader.isLinux(osName) && NativeLibraryLoader.X86.contains(osArch), 
			null, "libXkbDetectableAutoRepeat-linux.so");
	protected static final PlatformSupport LINUX64 = new PlatformSupport(
			(osName, osArch) -> NativeLibraryLoader.isLinux(osName) && NativeLibraryLoader.AMD64.contains(osArch), 
			null, "libXkbDetectableAutoRepeat-linux64.so");

	/**Applies the {@link XkbDetectableAutoRepeat} fix if the given window is an X11 window and ths current platform is supported.*/
	public static void apply(Display display, GLWindow window)
	{
		if(window.getUpstreamWidget() instanceof jogamp.newt.driver.x11.WindowDriver) {
			window.runOnEDTIfAvail(true, new Runnable() {
				@Override
				public void run() {
					String tmpDir = System.getProperty("java.io.tmpdir");
					File temp = new File(tmpDir, "XkbDetectableAutoRepeat");
					NativeLibraryLoader nut = new NativeLibraryLoader(XkbDetectableAutoRepeat.class, temp, false, LINUX, LINUX64);
					RecursiveLock lock = ((jogamp.newt.driver.x11.WindowDriver)window.getUpstreamWidget()).getLock();
					lock.lock();
					try {
						nut.extract();
						nut.load();
						System.out.println("Applying xkb detectable autorepeat (displayHandle="+display.getHandle()+")");
						if(display instanceof jogamp.newt.driver.x11.DisplayDriver)
							XkbDetectableAutoRepeat.XkbSetDetectableAutoRepeat(display.getHandle(), true);
					} catch(Exception err) {
						System.err.println("Warning: Failed to apply xkb detectable autorepeat: ");
						err.printStackTrace();
					} finally {
						lock.unlock();
					}
				}
			});
		}
	}
}
