package owg.lgm.util;

import java.util.Arrays;

import owg.util.stringescaper.CharEscaper;

public class DSGrid
{
	protected int width, height;
	protected Object[][] data;
	
	public DSGrid(int w, int h)
	{
		width = w;
		height = h;
		data = new Object[h][w];
		clear(0.0);
	}
	
	public void clear(String value)
	{
		for(int y = 0; y<height; y++)
			Arrays.fill(data[y], value);
	}
	public void clear(double value)
	{
		for(int y = 0; y<height; y++)
			Arrays.fill(data[y], value);	
	}
	
	public void set(int x, int y, String value)
	{
		if(x<0 || y<0 || x>=width || y>=height)
			return;
		data[y][x] = value;
	}
	public void set(int x, int y, double  value)
	{
		if(x<0 || y<0 || x>=width || y>=height)
			return;
		data[y][x] = value;
	}
	
	public void add(int x, int y, String value)
	{
		if(x<0 || y<0 || x>=width || y>=height)
			return;
		data[y][x] = getString(x, y).concat(value);
	}
	
	public void add(int x, int y, double value)
	{
		if(x<0 || y<0 || x>=width || y>=height)
			return;
		data[y][x] = getDouble(x, y)+value;
	}

	public String getString(int x, int y)
	{
		if(x<0 || y<0 || x>=width || y>=height)
			return "0";
		return String.valueOf(data[y][x]);
	}
	public double getDouble(int x, int y)
	{
		if(x<0 || y<0 || x>=width || y>=height)
			return 0.0;
		Object r = data[y][x];
		if(r instanceof Number)
			return ((Number) r).doubleValue();
		else
			return 0;
	}
	public int getInt(int x, int y)
	{
		if(x<0 || y<0 || x>=width || y>=height)
			return 0;
		Object r = data[y][x];
		if(r instanceof Number)
			return ((Number) r).intValue();
		else
			return 0;
	}
	
	protected static final CharEscaper ESC = new CharEscaper('\\', 'n', '\n', ':', ':');
	
	public String write()
	{
		StringBuilder sb = new StringBuilder();
		sb.append(width);
		sb.append(':');
		sb.append(height);
		sb.append(':');
		for(int y = 0; y<height; y++)
		{
			for(int x = 0; x<width; x++)
			{
				if(data[y][x] instanceof String)
				{
					sb.append('$');
					sb.append(ESC.escape((String)data[y][x]));
					sb.append(':');
				}
				else
				{
					sb.append('#');
					sb.append(data[y][x]);
					sb.append(':');
				}
			}
		}
		return sb.toString();
	}
	
	public void read(String str) throws IllegalArgumentException
	{
		String[] segments = str.split(":");
		//Skip empty
		int i = 0;
		while(i < segments.length && segments[i].isEmpty())
			i++;
		if(segments.length < i+2)
			throw new IllegalArgumentException("At least two non-empty lines required, got "+(segments.length-i));
		width = Integer.parseInt(segments[i]);
		i++;
		height = Integer.parseInt(segments[i]);
		i++;
		data = new Object[height][width];
		if(segments.length < i+width*height)
			throw new IllegalArgumentException("Required ("+width+"x"+height+") lines, got "+(segments.length-i));
		
		for(int y = 0; y<height; y++)
		{
			for(int x = 0; x<width; x++)
			{
				if(segments[i].startsWith("$"))
					data[y][x] = ESC.unescape(segments[i].subSequence(1, segments[i].length()));
				else if(segments[i].startsWith("#"))
					data[y][x] = Double.parseDouble(segments[i].substring(1));
				else
					throw new IllegalArgumentException("Unrecognized type of \""+segments[i]+"\"");
				i++;
			}
		}
	}
}
